#include <p18cxxx.h>
#include <sw_uart.h>
#include <delays.h>
#include <usart.h>
#include <stdlib.h>
#include <string.h>
#include <adc.h>
#include <math.h>
#include <stdio.h>

//NOTE- these values are included for refrence only, changing them may break the bootloader and 
//require a hardware programmer to fix.
#pragma config PLLDIV   = 4       // (16 MHz input)
#pragma config CPUDIV   = OSC1_PLL2
#pragma config USBDIV   = 2       // Clock source from 96MHz PLL/2
#pragma config FOSC     = HSPLL_HS
#pragma config FCMEN    = OFF
#pragma config IESO     = OFF
#pragma config PWRT     = OFF
#pragma config BOR      = ON
#pragma config BORV     = 3
#pragma config VREGEN   = ON
#pragma config WDT      = OFF
#pragma config WDTPS    = 32768
#pragma config MCLRE    = ON
#pragma config LPT1OSC  = OFF
#pragma config PBADEN   = OFF
#pragma config CCP2MX   = ON
#pragma config STVREN   = ON
#pragma config LVP      = OFF
//#pragma config ICPRT    = OFF       // Dedicated In-Circuit Debug/Programming
#pragma config XINST    = OFF       // Extended Instruction Set
#pragma config CP0      = OFF
#pragma config CP1      = OFF
#pragma config CP2      = OFF
//#pragma config CP3      = OFF
#pragma config CPB      = OFF
#pragma config CPD      = OFF
#pragma config WRT0     = OFF
#pragma config WRT1     = OFF
#pragma config WRT2     = OFF
//#pragma config WRT3     = OFF
#pragma config WRTB     = ON       // Boot Block Write Protection
#pragma config WRTC     = OFF
#pragma config WRTD     = OFF
#pragma config EBTR0    = OFF
#pragma config EBTR1    = OFF
#pragma config EBTR2    = OFF
//#pragma config EBTR3    = OFF
#pragma config EBTRB    = OFF

//Alias some human readable names to the hardware pins used
#define  RTS_UPLINK       PORTAbits.RA4
#define  STAT1            PORTBbits.RB6 
#define  STAT2            PORTBbits.RB7 
#define  GPS_RX           PORTCbits.RC0



//Prototypes
void DelayTXBitUART( void );     //needed for software UART
void DelayRXHalfBitUART( void ); //needed for software UART
void DelayRXBitUART( void );     //needed for software UART
char GPSRead( long *Time, long *Lattitude, long *Longitude, long *Altitude, int *H_Speed, int *V_Speed, int *Heading); //prototype for the Lassen IQ GPS module
char GPSStatus( char *Status );  //prototype to query the GPS status
void HardwareUARTInit( void );   //prototype to initialise the hardware UART
char USARTStatus ( void );
void SoftwareUARTInit( void );   //prototype to initialise the software UART
char ReadModem ( char *ModemData );
char ReadByteModem ( void );
char ReadCharModem( void );
void OutputLocation( void );



// **** Vector Remapping  ** /
//needed for workign with the USB bootloader
void YourHighPriorityISRCode();
void YourLowPriorityISRCode();

#define REMAPPED_RESET_VECTOR_ADDRESS			0x800
#define REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS	0x808
#define REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS	0x818
extern void _startup (void);        // See c018i.c in your C18 compiler dir
#pragma code REMAPPED_RESET_VECTOR = REMAPPED_RESET_VECTOR_ADDRESS
void _reset (void)
{
    _asm goto _startup _endasm
}
#pragma code REMAPPED_HIGH_INTERRUPT_VECTOR = REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS
void Remapped_High_ISR (void)
{
     _asm goto YourHighPriorityISRCode _endasm
}
#pragma code REMAPPED_LOW_INTERRUPT_VECTOR = REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS
void Remapped_Low_ISR (void)
{
     _asm goto YourLowPriorityISRCode _endasm
}
	

#pragma code HIGH_INTERRUPT_VECTOR = 0x08
void High_ISR (void)
{
	    _asm goto REMAPPED_HIGH_INTERRUPT_VECTOR_ADDRESS _endasm
}
#pragma code LOW_INTERRUPT_VECTOR = 0x18
void Low_ISR (void)
{
     _asm goto REMAPPED_LOW_INTERRUPT_VECTOR_ADDRESS _endasm
}
#pragma code
	
	
//These are your actual interrupt handling routines.
#pragma interrupt YourHighPriorityISRCode
void YourHighPriorityISRCode()
{
	//Check which interrupt flag caused the interrupt.
	//Service the interrupt
	//Clear the interrupt flag
	//Etc.
	
}	//This return will be a "retfie fast", since this is in a #pragma interrupt section 
#pragma interruptlow YourLowPriorityISRCode
void YourLowPriorityISRCode()
{
	//Check which interrupt flag caused the interrupt.
	//Service the interrupt
	//Clear the interrupt flag
	//Etc.
	
}	//This return will be a "retfie", since this is in a #pragma interruptlow section 





//vars  
char result;
char result2;
char WriteQuery[10] = "QL001;";
char address[10] = "001";
char good;
int loop;



void main(void) //start here
{
 WDTCON = 1;							//enable the watchdog timer
 LATA = 0b00000000;                     //all of the pins are set low at startup                                           
 LATB = 0b00100000;                     //except for GPS_TX, as the gps needs its 
 LATC = 0b00000000;                     //RX input driven high at all times
 RTS_UPLINK = 1;						//de-assert the rts line on the modem
 STAT1 = 1;								//turn on the yellow LED
 //STAT1 = 0;                             //turn on the green LED
 STAT2 = 1;								//turn on the red LED
 //STAT2 = 0;							  //turn on the blue LED
 TRISA = 0b00101111;                    //all of the pins are set as input at startup
 TRISB = 0b00000000;                    //except for TXD_UPLINK, OSC1, OSC2, GPS_TX
 TRISC = 0b11000001;                    //and the USB/USART ports
 HardwareUARTInit();					//initialise the UARTs
 SoftwareUARTInit();
 Delay1KTCYx(1);						//allow them to get ready to accept data
 printf( "\r\nK103-tx GPS Tracker Rev 0\r\n" ); 
 //main loop
 while(1)
  {
/*										//piece of code that just loops and data
										//recieved over the wireless modem back out
  while(ReadModem(data) != 0);			//useful for debugging
  putsUSART(data);
  Delay10KTCYx(100);
*/
 ClrWdt();								//reset the watchdog timer
										//if this function is not called for 2 mintutes
										//the processor will reset
								
										//start of main code loop
  result = ReadCharModem();				//wait for a new character from the modem
  if(result == WriteQuery[0])  			//check if it is the begining of the query
										//if it isn't get another one
  {
   for(loop = 1; WriteQuery[loop] != '\0'; loop++) //loop through the rest of the query
   {
    result2 = ReadCharModem();
    if( WriteQuery[loop] != result2 )        //if the character doesn't match then break
    {
     break;
    }
    if( ( WriteQuery[loop] == ';' ) & ( result2 == ';' ))  //if it detects the end of
    {										//the query output the current location
      OutputLocation();  
    }
   } 
   Delay10KTCYx(1);
  }

 }
}




//**********************************FUNCTIONS******************************************************


//HardwareUARTInit initialises the hardware USART init at 56700baud (for a 16mhz xtal)
//BAUD=CLKSPEED/(16*(SPBRG+1))
//NOTE: to close it, call CloseUSART()
//NOTE: you must have at least 200 TCYs between calling HardwareUARTInit and sending data to the uart
void HardwareUARTInit( void )
{
 OpenUSART ( 
  USART_TX_INT_OFF     &          //config the uart no tx interupt
  USART_RX_INT_OFF     &          //no rx interupt
  USART_ASYNCH_MODE    &          //anynch mode
  USART_EIGHT_BIT      &          //8 bits, no parity
  USART_CONT_RX        &          //rx all of the time
  USART_BRGH_LOW, 77 );          //the speed for 9600 baud and a 48MHz Fosc
 return;
}

//SoftwareUARTInit initialises the software UART code
//NOTE: to change the speed, you must edit DelayTXBitUART, DelayRXHalfBitUART, and DelayRXBitUART
void SoftwareUARTInit( void )
{
OpenUART();                        //configures the necessary pins to use the software UART
}
//***delay functions needed for software uart ***********************
//they are to set the software UART to run at 38400 baud transmit
//NOTE- these required some tweaking if you change the clock frequinecy
//delay for (((2*FOSC)/(4*BAUD))+1)/2)-12 cycles
void DelayTXBitUART(void)
{
Delay10TCYx ( 30 );
Delay1TCY();
return;
}
//delay for (((2*FOSC)/(8*BAUD))+1)/2)-9 cycles
void DelayRXHalfBitUART(void)
{
Delay10TCYx ( 15 );
return;
}
//delay for (((2*FOSC)/(4*BAUD))+1/2)-14 cycles
void DelayRXBitUART(void)
{
Delay10TCYx ( 28 );
return;
}
 
//GPSRead is used to read the time, location, and speed from the Lassen IQ GPS module
//It returns 0 if the read was sucessfull, -1 if the read times out
//possible causes for a read to fail include:
//the GPS was not ready to accept commands (true for about 2 seconds after power up)
//the GPS was not present
//the timing for the software USART was incorrect
//NOTE: You must call SoftwareUARTInit() before calling this function
char GPSRead( long *Time, long *Lattitude, long *Longitude, long *Altitude, int *H_Speed, int *V_Speed, int *Heading)
{
long GPSReady;
char time[6], latitude[11], longitude[12], altitude[8], h_speed[5], v_speed[6], heading[5], query [] = ">QLN<";
putsUART( query );                //send the query >QLN< to the GPS
                                  //the getsUART has no timeout, so we need our own
GPSReady = 0;                     //reset the timout counter to 0
while( GPS_RX )                   //the line will be high until the GPS responds, so once it is low we know the GPS is there
 {                                //but until it falls low we will loop for a few cycles and wait for it
 GPSReady = GPSReady + 1;         //each loop incrament the counter
 if( GPSReady > 10000)            //and if we loop too many times
  return -1;                      //assume the GPS isn't going to respond, and return -1
 }
getsUART(longitude, 9);           //the first 6 bytes that come back (>,R,L,N) we don't care about,
                                  // so we will temporarily store them in longitude, to be written over later
getsUART(time, 5);                //the next 5 bytes are the GPS time
time[5] = '\0';                   //null terminate the string
getsUART(longitude, 3);           //folowed by 3 bytes of fractional time, that we don't care about  
                                  // put it in longitude            
getsUART(latitude, 10);           //null terminate the string
latitude[10] = '\0';              //the next 10 bytes are the lattitude
getsUART(longitude, 11);          //followed by 11 bytes of longitude
longitude[11] = '\0';             //null terminate the string
getsUART(altitude, 7);            //and 7 bytes of alitude
altitude[7] = '\0';               //null terminate the string
getsUART(h_speed, 2);             //followed by 2 bytes of fractiona altitude which gets put into h_speed for now
getsUART(h_speed, 4);             //and 4 bytes of horisontal speed
h_speed[4] = '\0';                //null terminate the string
getsUART(v_speed, 5);             //and 5 bytes of vertical speed
v_speed[5] = '\0';                //null terminate the string
getsUART(heading, 4);             //and 4 bytes holding the heading.
heading[4] = '\0';                //null terminate the string
                                  // there is more data after this, but we will ingore it
*Time = atol( time );             //convert the strings from the GPS into integers, and store the result in the
*Lattitude = atol( latitude );    // location stored in the pointers passed when the function was called
*Longitude = atol( longitude );
*Altitude = atol( altitude );
*H_Speed = atoi( h_speed );
*V_Speed = atoi( v_speed );
*Heading = atoi( heading);
return 0;                         //return sucessfully
}

//GPSStatus is used to determins the status of the Lassen IQ GPS
//It returns -1 if the write fails, 0 if it is sucessfull
//NOTE: You must call SoftwareUARTInit() before calling this function
char GPSStatus( char Status[6] )
{
long GPSReady;
char query [] = ">QST<";
putsUART( query );                	//send the query >QST< to the GPS
                                  	//the getsUART has no timeout, so we need our own
GPSReady = 0;                     	//reset the timout counter to 0
while( GPS_RX )                   	//the line will be high until the GPS responds, so once it is low we know the GPS is there
 {                                	//but until it falls low we will loop for a few cycles and wait for it
 GPSReady = GPSReady + 1;         	//each loop incrament the counter
 if( GPSReady > 10000)            	//and if we loop too many times
  return -1;                      	//assume the GPS isn't going to respond, and return -1
 }
getsUART( Status, 4);           	//the first 4 bytes that come back (>,R,S,T) we don't care about,
                                  	// so we will temporarily store them in GPSStatus
getsUART( Status, 4);               //the next 4 bytes are the GPS status
Status[4] = '\0';                   //null terminate the string
return 0;                         	//return sucessfully
}
//ConvertToUTM is used to convert the longitude/lattitude coordinants produced by a GPS into easting/northing/zone format
//it assumes the input datum is WGS84, and outputs UTM (specifically NATOs UTM)  
//see http://www.uwgb.edu/dutchs/FieldMethods/UTMSystem.htm for more information
//you give it a pointer to longitude in *Easting (+ for east, - for west) and latitude in *Northing (+ for north, - for south), 1 count = 1e-7 degrees ie, 1231234567 for 123.1234567 degrees.
//it will change the values stored in the variables to northing/easting, and return the zone that you are in
//feeding it a 0 for longitude or lattitude will cause it to assume you are giving it empty data and return 0, and leave the data untouched
char ConvertToUTM( long *Easting, long *Northing)
{
 float lat, lon, sinlat, tanlat, rho, nu, s, p, coslat;
 char zone;
 if((*Easting == 0) || (*Northing == 0))  //if you feed it empty data just return empty data
  return 0;
 lon = *Easting / 572957795.1; //convert input value to radians
 lat = *Northing / 572957795.1;
 sinlat = sin(lat); //these functions are used a lot so it is faster to calculate them now
 coslat = cos(lat); 
 tanlat = sinlat/coslat;
 rho = 6335151.5 / pow(1-(0.006739497*sinlat*sinlat), 1.5);  //calculate the radius of curvature
 nu = 6378137.0 / pow(1-(0.006739497*sinlat*sinlat), .5);
 zone = (lon * 9.549296586 )+31; 	//calculate zone
 p = ((lon/.0174532925)-(zone*6.0-183))*.36;   //calculate the diffence in  
								//longitude from the centeral meridian to your location
 s = 6367449.15*lat-16038.43*sin(2*lat)+16.83261*sin(4*lat);  //calculate  meridional arc length
 *Easting = 500000+(nu*coslat*0.04846201*p+(coslat*coslat*	//calculate easting
   coslat*(1-tanlat*tanlat)+0.006739497*coslat*coslat)*120*p*p*p);
 *Northing = s*0.9996+nu*sinlat*coslat*0.00117475*p*p;	//calculate northing
 if( lat < 0)
  *Northing += 10000000;
 return zone;			//return the zone
}


//USARTStatus is used to determine if there is data available in the hardware USART FIFO buffer
//It returns 0 if there is a character waiting to be read, -1 if no character is found
//It checks the buffer several times over the course of several bit periods, and will wait for a bit to
//arrive before failing.
//NOTE- USARTStatus DOES NOT active the RTS line on the modem, the user must pull down 
//the RTS line before calling USARTStatus
char USARTStatus ( void )
{
 char USARTloop = 0;
 for (USARTloop; USARTloop < 50; USARTloop++ ) 	//check for data 50 times
 {
  if( DataRdyUSART() )							//if there is a character waiting
												//in the uart
  {
   return 0;									//return sucessful
  }
  Delay1KTCYx(2);
 }
 return -1;										//esle return a fail
}

//ReadModem is used to read the data available in the MaxStream wireless modem
//It returns 0 if there was data waiting to be read, -1 if no data was found
//NOTE- any data after the first ~100 characters will not be read, and needs
// to wait until the next time the function is called to be read. 
char ReadModem ( char ModemData[20] )
{
 char ReadModemLoop = 0;
 ModemData[0] = '\0';             //clear out any old data 
 if( USART_Status.OVERRUN_ERROR )  //if the hardware USART's buffer has been overrun, reset it
  {
   CloseUSART();
   HardwareUARTInit();
   Delay1KTCYx(1);
  }
 RTS_UPLINK = 0;                  //bring the RTS line of the modem down, which tells it to send data
 if( USARTStatus() == 0 )         //if it sends any data out, start a loop to collect it
 {
  for(ReadModemLoop; ( USARTStatus() == 0 ); ReadModemLoop++ )
  {
   if( ReadModemLoop > 1 )		  //if the buffer is almost full, bring up the RTS line
    RTS_UPLINK = 1;				  //this value can be tweaked to change the number of 
								  //characters recieved, make sure that ModemData is big enough to hold it
   ModemData[ReadModemLoop] = ReadUSART(); //create a string holding the data sent
  }
  ModemData[ReadModemLoop] = '\0';   //add a null terminator
  RTS_UPLINK = 1;                  //bring the RTS line back high to hold data until the next time it is needed
  return 0  ;                      //and return sucessfully
 }
 else                              //if no data was recieved
 {
 RTS_UPLINK = 1;                   //bring the RTS line up to hold any data that might come later
 return -1;                        //and return a faulure
 }
}

//ReadByteMoem is used to read one byte out of the MaxStream wireless modem
//It returns the next byte available from the MaxStream modem, or a null ('\0') if 
//there was no new data available
char ReadByteModem ( void )
{
 static char ModemDataBuffer[20] = {'\0'};
 static char ModemDataIndex = 0;
 char ModemReadResult;
 if( ModemDataBuffer[ModemDataIndex] == '\0' ) //if the current character is a null get new data
 {
  ModemReadResult = ReadModem( ModemDataBuffer );
  if( ModemReadResult != 0) 				 //if there wasn't data
  {
   ModemDataBuffer[0] = '\0';
   return '\0';                            //if there was no new data return a null
  }
  if(ModemReadResult == 0)                	//if there was data return the first byte
  {
   ModemDataIndex = 1;                    	//and reset the current character location to 1 (since we just read location 0)
   return ModemDataBuffer[0];
  }
 }
 else										//if the current character wasn't a null
 { 
  ModemDataIndex = ModemDataIndex + 1;		//incrament the location counter
  return ModemDataBuffer[ModemDataIndex - 1];	//and return the character at the current location
 } 
}

//ReadByteMoem is used to read one byte out of the MaxStream wireless modem
//It waits for a byte of data to be available from the MaxStream modem and 
//returns returns it
char ReadCharModem( void )
{
 char CharBuffer = '\0';
 while(1)							//loop until the program returns sucessfully
 {
  CharBuffer = ReadByteModem();		//read the last byte from the modem
  if(CharBuffer != '\0')			//if it isn't a null
   return CharBuffer;				//return it
 } 
}

//OutputLocation is used to print the current coordinants of the tracker
//it prints a line 'address	northing	easting	zone	gps status'
void OutputLocation( void )
{
 char good2;
 long GPS_Time;
 long GPS_Lattitude;
 long GPS_Longitude;
 long GPS_Altitude;
 int GPS_Horisontal_Speed;
 int GPS_Vertical_Speed; 
 int GPS_Heading;
 char GPS_Status[6];
 char zone;
 long *pGPS_Time = &GPS_Time;
 long *pGPS_Lattitude = &GPS_Lattitude;
 long *pGPS_Longitude = &GPS_Longitude;
 long *pGPS_Altitude = &GPS_Altitude;
 int *pGPS_Horisontal_Speed = &GPS_Horisontal_Speed;
 int *pGPS_Vertical_Speed = &GPS_Vertical_Speed;
 int *pGPS_Heading = &GPS_Heading;
 pGPS_Time = &GPS_Time;
 pGPS_Lattitude = &GPS_Lattitude;
 pGPS_Longitude = &GPS_Longitude;
 pGPS_Altitude = &GPS_Altitude;
 pGPS_Horisontal_Speed = &GPS_Horisontal_Speed;
 pGPS_Vertical_Speed = &GPS_Vertical_Speed;
 pGPS_Heading = &GPS_Heading;
  GPSStatus( GPS_Status );		//check the GPS status
  good = GPSRead( pGPS_Time, pGPS_Lattitude, pGPS_Longitude, pGPS_Altitude, pGPS_Horisontal_Speed, pGPS_Vertical_Speed, pGPS_Heading );
								//read the longitude/lattitude out of the GPS
  if( good == 0 )				//if the reads were sucessful
   {
   zone = ConvertToUTM( pGPS_Longitude, pGPS_Lattitude); //convert LLA to UTM format
   printf( "%s\t%ld\t%ld\t%d\r\n%s\r\n", address, GPS_Lattitude, GPS_Longitude, zone, GPS_Status);
								//and print the result
   }
  else
   printf( "GPS Read Error!\r\n" );  //otherwise print an error
return;
}
